#Copyright (c) 2008 Vitor Bosshard
#This program licensed under the GPL. See license.txt for details.
#
#Tested under:
#Python 2.6.1
#Pygame 1.8.1.win32-py2.5
#-----------------------------------------------------------------

import random, os, sys, marshal, re, copy
from pyflam3 import *
from math import *

try:
    import pygame
except ImportError:
    raise
    print("Warning: Couldn't import pygame.")
    # Ugly hack or clever trick? you decide!
    pygame = False
    

class ParsingError(Exception):
    pass


class Xaoslist(list):
    """ A list which returns 1 for unassigned items, and pads the list when
    necessary to store values in their correct locations."""

    def __repr__(self):
        return str(self[:])

    def __init__(self,parent,lst):
        # Assigning a bound method would create problems with deepcopy
##        self._len = parent.parent.xform.__len__
        self.parent = parent
        self.xform = parent.parent.xform
        lst.extend(1 for i in range(100-len(lst)))
        list.__init__(self,lst)

    def __len__(self):
##        return self._len()
        if self.parent is self.parent.parent.final:
            return 0
        return len(self.xform)

    def __iter__(self):
        return (self[i] for i in range(len(self)))
        
    def __getitem__(self,pos):
        if abs(pos) > len(self)-1:
            raise IndexError
        return list.__getitem__(self,pos)

    def __getslice__(self,pos,pos2):
        if (pos<0) or (pos2<0):
            raise NotImplementedError, "Negative slicing is not supported"
        if pos2 > len(self): pos2 = len(self)
        return list.__getslice__(self,pos,pos2)
    
    def __setitem__(self,pos,val):
        if abs(pos) > len(self)-1:
            raise IndexError
        list.__setitem__(self,pos,val)
      
    def __setslice__(self,pos,pos2,val):
        if (pos<0) or (pos2<0):
            raise NotImplementedError, "Negative slicing is not supported"
        if pos2 > len(self): pos2 = len(self)
        if len(val) != pos2-pos:
            raise IndexError, "Assigned slice has incorrect lenght"
        list.__setslice__(self,pos,pos2,val)

    def get_list(self):
        lst = list.__getslice__(self,0,len(self))
        for i in reversed(lst):
            if i != 1: break
            lst.pop()
        return lst
            

class Xform(object):
    """Container for transform parameters."""
    
    def __repr__(self):
        if self is self.parent.final:
            return "<finalxform>"
        else:
            return "<xform %d>" % self.parent.xform.index(self)
      
    def __getattr__(self,v):
        try:
            return object.__getattribute__(self,v)
        except AttributeError:
            return 0

    def __setattr__(self,name,v):
        """Protects special attributes from being irrecoverably overwritten"""

        if name is 'chaos' and type(v) is not Xaoslist:
            raise TypeError("Can't override chaos with %s" %type(v))
        elif name is 'post' and type(v) is not PostXform:
            raise TypeError("Can't override post xform with %s" %type(v))
        elif v == 0 and name is not 'color':
            try:
                delattr(self,name)
            except AttributeError:
                pass
        else:
            object.__setattr__(self,name,v)
    
    def __init__(self,parent,values):
        self.parent = parent
        for name,val in values:
            if name is 'chaos':
                self.chaos = Xaoslist(self,val)
            elif name is 'post':
                self.post = PostXform(self,(("coefs",val),))              
            else:
                setattr(self,name,val)

        # Convert from "screen" to "complex plane" format
        self.d = -self.d
        self.b = -self.b
        self.f = -self.f
        
        # Create default values. Subclasses ignore this.
        if type(self) is Xform:
            if not self.chaos:
                self.chaos = Xaoslist(self,[])
            if not self.post:
                self.post = PostXform(self,(("coefs",[1,0,0,1,0,0]),))    

    def _get_coefs(self):
        return self.a,self.d,self.b,self.e,self.c,self.f

    def _set_coefs(self,v):
        self.a,self.d,self.b,self.e,self.c,self.f = v

    coefs = property(_get_coefs,_set_coefs)

    def get_screen_coefs(self):
        """Creates a list of coefs in "screen" notation."""
        return self.a,-self.d,-self.b,self.e,self.c,-self.f

#----------------------------------------------------------------------
    def _set_position(self,v1,v2=None):
        if v2 is None: v1, v2 = v1
        self.c = v1
        self.f = v2
        
    def _get_position(self):
        return self.c, self.f
    
    position = property(_get_position,_set_position)

    def move_position(self,v1,v2=None):
        if v2 is None: v1, v2 = v1       
        self.c += v1
        self.f += v29

#----------------------------------------------------------------------
    def _set_x(self,v1,v2=None):
        if v2 is None: v1, v2 = v1
        self.a  = v1 - self.c
        self.d  = v2 - self.f
        
    def _get_x(self):
        return self.a + self.c, self.d + self.f
    
    x = property(fget=_get_x,fset=_set_x)

    def move_x(self,v1,v2=None):     
        if v2 is None: v1, v2 = v1  
        self.a += v1
        self.d += v2

#----------------------------------------------------------------------
    def _set_y(self,v1,v2=None):
        if v2 is None: v1, v2 = v1
        self.b  = v1 - self.c
        self.e  = v2 - self.f
        
    def _get_y(self):
        return self.b + self.c, self.e + self.f

    y = property(fget=_get_y,fset=_set_y)

    def move_y(self,v1,v2=None):     
        if v2 is None: v1, v2 = v1 
        self.b += v1
        self.e += v2

#----------------------------------------------------------------------
    def _set_o(self,v1,v2=None):
        if v2 is None: v1, v2 = v1
        self.a += self.c - v1
        self.d += self.f - v2
        self.b += self.c - v1
        self.e += self.f - v2
        self.c  = v1
        self.f  = v2

    def _get_o(self):
        return self.c, self.f
    
    o = property(fget=_get_o,fset=_set_o)

    def move_o(self,v1,v2=None):
        if v2 is None: v1, v2 = v1
        self.a -= v1
        self.d -= v2
        self.b -= v1
        self.e -= v2
        self.c += v1
        self.f += v2

#----------------------------------------------------------------------
    def scale(self,v):
        self.a *= v
        self.d *= v
        self.b *= v
        self.e *= v

    def rotate(self,deg,pivot="local"):
        if pivot is "local":
            # Get the absolute angle each triangle leg will have after rotating. 
            # Atan2 puts the result into the proper quadrant automatically.
            angle   = radians(deg)
            x_angle = atan2(self.a,self.d) + angle
            y_angle = atan2(self.b,self.e) + angle
            
            # Get the length of the triangle leg, and reconstruct the
            # coordinate pair with the new angle.
            x_leg  = hypot(self.a,self.d)
            self.a = sin(x_angle) * x_leg
            self.d = cos(x_angle) * x_leg
            
            y_leg  = hypot(self.b,self.e)
            self.b = sin(y_angle) * y_leg
            self.e = cos(y_angle) * y_leg
        else:
            self.rotate(deg)
            self.orbit(deg,pivot)

    def orbit(self,deg,pivot=(0,0)):
        """Orbits the transform around a fixed point without rotating it."""
        hor = self.c - pivot[0]
        ver = self.f - pivot[1]   
        angle  = atan2(hor,ver) + radians(deg)
        
        vector = hypot(hor,ver)
        self.c = pivot[0] + sin(angle) * vector
        self.f = pivot[1] + cos(angle) * vector

    def copy(self):
        self.parent.xform.append(copy.deepcopy(self))

    def delete(self):
        if self.parent.final is self:
            self.parent.final = None            
        else:
            self.parent.xform.remove(self)


class PostXform(Xform):
    _allowed = ['coefs','parent','a','b','c','d','e','f']
    
    def __repr__(self):
        xform = repr(self.parent)
        return "<post-%s" % xform[1:]

    def __setattr__(self,name,v):
        if name not in self._allowed:
            raise AttributeError, 'Can\'t assign "%s" to %s' %(name,self)
        object.__setattr__(self,name,v)

    def copy(self):
        raise TypeError, "Can't copy a post transform"

    def delete(self):
        raise TypeError, "Can't delete a post transform"    


class Flame(object):
    re_flame  = re.compile(r'<flame .*?</flame>',re.DOTALL)
    re_header = re.compile(r'<flame .*?>')
    re_xform  = re.compile(r'<[a-zA-Z]*xform .*?/>')
    re_attr   = re.compile(r'[^ ]*?=".*?(?=")') # Works for xforms and header  
    re_grad   = re.compile(r'[0-9A-F]{6}(?=[0-9A-F]*.?$)',re.MULTILINE) 

    def __init__(self,filename="",flamename=""):
        self.flamename = flamename
        self.filename = os.path.join(sys.path[0],"parameters",filename)
        self.gradient = []
        self.xform = []
        self.pixels = []

        if filename:
            self.from_file(self.filename,self.flamename)

            self._default_xform = dir(Xform(self,(("coefs",[1,1,1,1,1,1]),)))
            self._default_flame = dir(Flame())
        else:
            self._default_flame = None
            self._default_xform = None
            self.final  = None
            self.scale  = None

    def from_file(self,filename,flamename):
        """Retrieves all flame data from a flame file."""

        f = open(filename,"r")
        flames = self.re_flame.findall(f.read())
        f.close()
        
        f = False
        for flame in flames:
            # if flamename is not given, this just takes the first flame.
            if 'name="%s' %flamename in flame:
                f = flame ; break
        if not f:
            raise NameError, 'Flame "%s" not found' %flamename
            
        # Record the header data
        for string in self.re_attr.findall(self.re_header.search(f).group()):
            name, val = string.split('="')
            # Convert value to the appropriate type
            try:
                if " " in val: val = map(float,val.split())
                else:          val = float(val)
            except:
                pass   # Keep as string
            
            setattr(self,name,val)

        # Set required attributes that might be missing in the file.                       
        if not hasattr(self,"zoom"):
            self.zoom = 0
        if not hasattr(self,"center"):
            self.center = [0,0]
        if not hasattr(self,"angle"):
            self.angle = 0
        self.scale = self.scale * 100 / self.size[0]
        self.final = None
            
        # Create the gradient     
        for i in self.re_grad.findall(f):
            self.gradient.append((int(i[0:2],16),
                                  int(i[2:4],16),
                                  int(i[4:6],16)))
        if len(self.gradient) != 256:
            raise ParsingError("Palette data unreadable")

        mask = [ 6,12, 6,
                12,64,12,
                 6,12, 6]
        for r,g,b in self.gradient:
            pixel = pygame.Surface((3,3),pygame.SRCALPHA)
##            pixel.fill(color)
##            pixel.set_alpha(128)                
            for i in range(9):
                pixel.set_at((i%3,i/3),[r,g,b,mask[i]])
            self.pixels.append(pixel)
                
        # Create the Xform objects
        special = ["coefs","chaos","post"]
        for xform in self.re_xform.findall(f):
            parameters = []
            for string in self.re_attr.findall(xform):
                name, val = string.split('="')
                try:
                    if name in special: param = name, map(float,val.split())
                    else:               param = name, float(val)
                except:
                    param = name, val
                    
                parameters.append(param)

            x = Xform(self,parameters)              

            # Assign the xform to the correct location, 
            if x.weight:
                self.xform.append(x)
            elif not self.final:
                self.final = x
            else:
                raise ParsingError("More than one final xform found")

    def to_string(self,include_details=True):
        """Extracts parameters from a Flame object and converts them into string format."""

        # Write the flame header
        if include_details:
            string = '<flame '
            for name in dir(self):
                if name in self._default_flame: continue
                val = getattr(self,name)
                if   type(val) is list or type(val) is tuple:
                    #Convert round numbers to integer.
                    param = name," ".join(str(i).rstrip('0').rstrip('.') for i in val)                                                                               
                else:
                    param = name,val
                string += '%s="%s" ' %param
            string += 'scale="%s" >' %(self.scale * self.size[0] / 100)
        else:
            string = '<flame name="fr0st" >'

        # Make each xform
        xformlist = self.xform[:]
        if self.final:
            xformlist.append(self.final)
        for xform in xformlist:
            if xform is self.final:
                xform_str = '\n   <finalxform '
            else:
                xform_str = '\n   <xform '
##            xform_str = '\n   <xform '

            for name in dir(xform):
                if name in self._default_xform: continue
                xform_str += '%s="%s" ' %(name,getattr(xform,name))         
            coef_str = 'coefs="%s %s %s %s %s %s" ' % xform.get_screen_coefs()

            # Write the post xform only if it isn't neutral.
            post = xform.post.get_screen_coefs()
            if post != (1,0,0,1,0,0):
                post_str = 'post="%s %s %s %s %s %s" ' % post
            else:
                post_str = ''

            # Write the chaos values. get_list() has output identical to Apo.
            if xform.chaos:
                xaos = xform.chaos.get_list()
                if xaos:
                    chaos_str = 'chaos="%s" />' %" ".join(map(str,xaos))
                else:
                    chaos_str = '/>'
            else:
                chaos_str = '/>'
                
            string += xform_str + coef_str + post_str + chaos_str
        
        # Make the gradient
        if include_details:
            grad_str = '\n   <palette count="256" format="RGB">'
            for i,color in enumerate(self.gradient):
                if not i%8: grad_str += "\n      "
                grad_str += '%02X%02X%02X' %color

            return string + grad_str + '\n   </palette>\n</flame>\n'
        else:
            return string + '</flame>'

    def create_final(self):
        if self.final: return
        self.final = Xform(self,(("coefs",[1,0,0,1,0,0]),
                                 ("linear",1),
                                 ("color",0)))

    def add_xform(self):
        self.xform.append(Xform(self,(("coefs",[1,0,0,1,0,0]),
                                      ("linear",1),
                                      ("color",0))))

    def _get_rotate(self):
        return degrees(self.angle)

    def _set_rotate(self,v):
        self.angle = radians(v)

    rotate = property(_get_rotate,_set_rotate)
    

class Display(object):
    
    def _get_size(self):
        return self.width,self.height

    def _set_size(self,v):
        if hasattr(self,'screen'):
            raise AttributeError, "Screen size can't be changed once pygame is initialised."
        self.width  = v[0]
        self.height = v[1]

    size = property(fget=_get_size,fset=_set_size)

    def _get_alpha(self):
        return self.background.get_alpha()

    def _set_alpha(self,v):
        self.background.set_alpha(v)
    
    background_alpha = property(_get_alpha,_set_alpha)

    width = 640
    height = 480
    quality = 9000
    rc = RandomContext()
    xform_distrib = ARRAY(c_ushort,10000)()  # must be at least 10000
    iteration = 0
    running = False
    _tick = 0
    _frames_rendered = 0
    keydown = []

    def load_flame(self,filename,flamename=""):
        self.flame = Flame(filename,flamename)
        return self.flame
    
    def __init__(self):
        if pygame:
            self.background = pygame.Surface(self.size)
            self.background.set_alpha(8)
        
        self.load_flame("samples.flame")

    def init(self):
        #Init Pygame
        if not pygame: return
        self.running = True
        pygame.init()
        pygame.display.set_caption('Fractal Fr0st')
        alpha = self.background_alpha
        self.background = pygame.Surface(self.size)
        self.background_alpha = alpha
        Array = ARRAY(c_double,self.quality*4)
        self.samples = Array(flam3_random_isaac_11(self.rc),
                             flam3_random_isaac_11(self.rc),
                             0,
                             1)
        
    def iterate(self,flame):
        # Local variables
        l_screen = self.screen
        l_samples = self.samples
        c0,c1 = flame.center
        
        # Precalc some stuff used to project the points to the screen.
        cpx = self.width/4  # 1 complex plane unit, in pixels.
        cpxcos = cos(flame.angle) * cpx
        cpxsin = sin(flame.angle) * cpx
        offset0 = self.width/2 -1  # the -1 compensates for 3x3 pixelsize.
        offset1 = self.height/2 -1

        # Call flam3 functions
        genome = Genome.from_string(flame.to_string(False))[0]
     
        flam3_create_xform_distrib(byref(genome),
                                   cast(self.xform_distrib,POINTER(c_ushort)))

        prepare_xform_fn_ptrs(byref(genome),
                              byref(self.rc))

        flam3_iterate(byref(genome),
                      self.quality,
                      20,
                      cast(l_samples,POINTER(c_double)),
                      cast(self.xform_distrib,POINTER(c_ushort)),
                      byref(self.rc))

        # Draw the results
        for i in range(0,self.quality*4,4):
            x,y,c = l_samples[i:i+3]
            # Rotate
            x -= c0
            y -= c1
            coord = (x*cpxcos + y*cpxsin + offset0,
                     y*cpxcos - x*cpxsin + offset1)            

            try:
                l_screen.blit(flame.pixels[int(c*256)],coord)
            except:
                # The point is out of drawable range. It's cheaper to ignore
                # this than check for it at every iteration
                pass
            
        self.iteration += 1

    def render(self):
        if not self.running: return

        # Creates the screen the first time this function is called
        if not hasattr(self,"screen"):
            self.screen = pygame.display.set_mode(self.size)
            self._first_tick = pygame.time.get_ticks()
            self._tick = self._first_tick
        
        # Event handling is put here to simplify scripts
        self.keydown[:] = []
        for e in pygame.event.get():
            if e.type == pygame.KEYDOWN:
                self.keydown.append(e.key)
                if e.key == pygame.K_ESCAPE:
                    pygame.quit()
                    sys.exit()
            elif e.type == pygame.QUIT:
                pygame.quit()
                sys.exit()

        self.iterate(self.flame)
        self.screen.blit(self.background, (0,0))
        if pygame.time.get_ticks() > self._tick + 50: # sets display refresh rate to 20 FPS                 
            pygame.display.flip()
            self._tick += 50
            self._frames_rendered += 1

##        if pygame.time.get_ticks() > 1000 + self._first_tick:
##            print self.iteration * self.quality
##            print "%.1f%% of frames rendered" %(100.*(self._frames_rendered)/self.iteration)
##            pygame.quit()
##            a = raw_input('\nPress Enter to exit')
##            sys.exit()

