#Copyright (c) 2008 Vitor Bosshard
#This program licensed under the GPL. See license.txt for details.
#
#Tested under:
#Python 2.6.1
#Pygame 1.8.1.win32-py2.5
#-----------------------------------------------------------------


#This code sets up the namespace in which fr0st scripts run

import os, sys, marshal, copy, random
from math import *
sys.dont_write_bytecode = False

import fr0stlib
from fr0stlib import Xform, Flame
from pyflam3.variations import variations

try:
    import pygame
    from pygame import (KMOD_ALT,KMOD_CAPS,KMOD_CTRL,KMOD_LALT,KMOD_LCTRL,
KMOD_LMETA,KMOD_LSHIFT,KMOD_META,KMOD_MODE,KMOD_NONE,KMOD_NUM,KMOD_RALT,
KMOD_RCTRL,KMOD_RMETA,KMOD_RSHIFT,KMOD_SHIFT,K_0,K_1,K_2,K_3,K_4,K_5,K_6,K_7,
K_8,K_9,K_AMPERSAND,K_ASTERISK,K_AT,K_BACKQUOTE,K_BACKSLASH,K_BACKSPACE,K_BREAK,
K_CAPSLOCK,K_CARET,K_CLEAR,K_COLON,K_COMMA,K_DELETE,K_DOLLAR,K_DOWN,K_END,
K_EQUALS,K_ESCAPE,K_EURO,K_EXCLAIM,K_F1,K_F10,K_F11,K_F12,K_F13,K_F14,K_F15,
K_F2,K_F3,K_F4,K_F5,K_F6,K_F7,K_F8,K_F9,K_FIRST,K_GREATER,K_HASH,K_HELP,K_HOME,
K_INSERT,K_KP0,K_KP1,K_KP2,K_KP3,K_KP4,K_KP5,K_KP6,K_KP7,K_KP8,K_KP9,
K_KP_DIVIDE,K_KP_ENTER,K_KP_EQUALS,K_KP_MINUS,K_KP_MULTIPLY,K_KP_PERIOD,
K_KP_PLUS,K_LALT,K_LAST,K_LCTRL,K_LEFT,K_LEFTBRACKET,K_LEFTPAREN,K_LESS,K_LMETA,
K_LSHIFT,K_LSUPER,K_MENU,K_MINUS,K_MODE,K_NUMLOCK,K_PAGEDOWN,K_PAGEUP,K_PAUSE,
K_PERIOD,K_PLUS,K_POWER,K_PRINT,K_QUESTION,K_QUOTE,K_QUOTEDBL,K_RALT,K_RCTRL,
K_RETURN,K_RIGHT,K_RIGHTBRACKET,K_RIGHTPAREN,K_RMETA,K_RSHIFT,K_RSUPER,
K_SCROLLOCK,K_SEMICOLON,K_SLASH,K_SPACE,K_SYSREQ,K_TAB,K_UNDERSCORE,K_UNKNOWN,
K_UP,K_a,K_b,K_c,K_d,K_e,K_f,K_g,K_h,K_i,K_j,K_k,K_l,K_m,K_n,K_o,K_p,K_q,K_r,
K_s,K_t,K_u,K_v,K_w,K_x,K_y,K_z)
except ImportError:
    pass

def exit():
    try:     pygame.quit()
    finally: raise SystemExit

quit = exit

display = fr0stlib.Display()
flame = display.flame
load_flame = display.load_flame
keydown = display.keydown


try:
    display.init()
except:
    print("Warning: Can't initialize pygame, render requests will be ignored.")
#-------------------------------------------------------------------------------


def add_xform():
    flame.xform.append(Xform(flame,(("coefs",[1,0,0,1,0,0]),),))

def clear():
    flame.xform = []
    flame.final = None

    

class Recorder():
    """Records, loads and renders a sequence of frames."""
    # Warning: This class has not been updated and is for use only with
    # classic_iterate
    history=[]

    def record(self,xforms):
        """Converts the current state of all xforms into a list and adds it to self.history."""   
        l_xforms = [] 
        for xform in xforms: 
            xform = xform.get_screen_coefs() + (xform.weight,xform.color)
            l_xforms.append(xform)
        self.history.append(l_xforms)

    def save_frost(self,flame,filename):
        f = file(os.path.join(sys.path[0],"parameters",filename),"w+b")
        data = self.history,flame.gradient,flame.zoom,flame.center
        marshal.dump(data,f)

    def load_frost(self,flame,filename,convert=True):
        """Loads parameters from a file, and converts them back to a regular format unless convert is False."""
        f = open(os.path.join(sys.path[0],"parameters",filename),"r+b")
        data = marshal.load(f)
        self.history,flame.gradient,flame.zoom,flame.center = data
        if convert: history = self.convert_frost(data[0])

    def convert_frost(self,history):
        """Converts a loaded file into a Frost readable format."""
        temp_history = []
        for frame in history:
            temp_frame = []
            for xform in frame:
                values = [("coefs",xform[0:6]),("weight",xform[6]),("color",xform[7])]
                temp_frame.append(Xform(flame,values))
            temp_history.append(temp_frame)
        return temp_history
    
    def save_image(self,display,ext="png"):
        img = display.screen
        i = display.iteration
        pygame.image.save(img, os.path.join(sys.path[0],"images","frame%04d." %i + ext))

    def interpolate(self,number):
        """Interpolates frames into the flame history.

        For coloring to work properly, 'number' must be a power of 2.
        The number of output frames is 2n-1."""

        if (number % 2 == 0) and (number > 0):
            interpolate(number/2)
            
            newhistory = []
            for xforms,xforms2 in zip(self.history[:-1],self.history[1:]):
                extra_xforms = []
                for xform,xform2 in zip(xforms,xforms2):
                    extra_xform = []
                    for coef,coef2 in zip(xform,xform2):
                        extra_coef = coef*0.5 + coef2*0.5
                        extra_xform.append(extra_coef)
                    extra_xforms.append(extra_xform)
                newhistory.append(xforms)
                newhistory.append(extra_xforms)
            newhistory.append(self.history[-1]) #adds the last frame without interpolating.
            self.history = newhistory

        elif number != 1:
            print "Warning: x is not a power of 2. Press enter to quit." ; input()


class Exporter():
    """Creates .flame parameters from fr0st's flame object.""" 
    _flames = []

    def flame_to_buffer(self,flame,display):
        """Accumulates flame parameters in a string buffer."""
        self._flames.append(flame.to_string())

    def save_flame(self,flame,display):
        """Creates a .flame file containing a single flame."""
        self._tempflames = self._flames
        self._flames = []
        self.flame_to_buffer(flame,display)
        self.save_flame_buffer("flame.flame")
        self._flames = self._tempflames
        del self._tempflames
        
    def save_flame_buffer(self,filename):
        """Writes the string buffer to a .flame file."""
        f = open(filename,"w")
        f.write("""<Flames name="Fr0st Batch">\n""")
        f.write("".join(self._flames))
        f.write("""</Flames>""")
        f.close()

recorder      = Recorder()
record        = recorder.record
save_frost    = recorder.save_frost
load_frost    = recorder.load_frost
convert_frost = recorder.convert_frost
save_image    = recorder.save_image
interpolate   = recorder.interpolate

exporter          = Exporter()
flame_to_buffer   = exporter.flame_to_buffer
save_flame        = exporter.save_flame
save_flame_buffer = exporter.save_flame_buffer


del Exporter,Recorder

